package com.cursedcauldron.wildbackport.common.blocks;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Random;
import java.util.Set;
import java.util.function.Supplier;
import net.minecraft.class_156;
import net.minecraft.class_1923;
import net.minecraft.class_1936;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2382;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2509;
import net.minecraft.class_2680;
import net.minecraft.class_3218;
import net.minecraft.class_3419;
import net.minecraft.class_3532;
import net.minecraft.class_6019;
import net.minecraft.class_638;
import net.minecraft.class_6862;
import com.cursedcauldron.wildbackport.WildBackport;
import com.cursedcauldron.wildbackport.client.particle.SculkChargeParticleOptions;
import com.cursedcauldron.wildbackport.client.registry.WBParticleTypes;
import com.cursedcauldron.wildbackport.client.registry.WBSoundEvents;
import com.cursedcauldron.wildbackport.common.tag.WBBlockTags;
import com.cursedcauldron.wildbackport.common.utils.DirectionUtils;
import com.cursedcauldron.wildbackport.common.utils.ModUtils;
import com.cursedcauldron.wildbackport.common.utils.ParticleUtils;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import com.mojang.serialization.Codec;
import com.mojang.serialization.Dynamic;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;
import it.unimi.dsi.fastutil.objects.ObjectArrayList;
import org.jetbrains.annotations.Nullable;

public class SculkSpreadManager {
    final boolean isWorldGen;
    private final class_6862<class_2248> replaceableBlocks;
    private final int extraBlockChance;
    private final int maxDistance;
    private final int spreadChance;
    private final int decayChance;
    private List<Cursor> cursors = new ArrayList<>();

    public SculkSpreadManager(
        boolean isWorldGen,
        class_6862<class_2248> replaceableBlocks,
        int extraBlockChance,
        int maxDistance,
        int spreadChance,
        int decayChance
    ) {
        this.isWorldGen = isWorldGen;
        this.replaceableBlocks = replaceableBlocks;
        this.extraBlockChance = extraBlockChance;
        this.maxDistance = maxDistance;
        this.spreadChance = spreadChance;
        this.decayChance = decayChance;
    }

    public static SculkSpreadManager create() {
        return new SculkSpreadManager(false, WBBlockTags.SCULK_REPLACEABLE, 10, 4, 10, 5);
    }

    public static SculkSpreadManager createWorldGen() {
        return new SculkSpreadManager(
            true, WBBlockTags.SCULK_REPLACEABLE_WORLD_GEN, 50, 1, 5, 10
        );
    }

    public class_6862<class_2248> getReplaceableBlocks() {
        return this.replaceableBlocks;
    }

    public int getExtraBlockChance() {
        return this.extraBlockChance;
    }

    public int getMaxDistance() {
        return this.maxDistance;
    }

    public int getSpreadChance() {
        return this.spreadChance;
    }

    public int getDecayChance() {
        return this.decayChance;
    }

    public boolean isWorldGen() {
        return this.isWorldGen;
    }

    public void clearCursors() {
        this.cursors.clear();
    }

    public void readTag(class_2487 tag) {
        if (tag.method_10573("cursors", 9)) {
            this.cursors.clear();
            List<Cursor> cursors
                = Cursor.CODEC.listOf()
                      .parse(new Dynamic<>(class_2509.field_11560, tag.method_10554("cursors", 10)))
                      .resultOrPartial(WildBackport.LOGGER::error)
                      .orElseGet(ArrayList::new);
            int size = Math.min(cursors.size(), 32);

            for (int i = 0; i < size; i++) {
                this.addCursor(cursors.get(i));
            }
        }
    }

    public void writeTag(class_2487 tag) {
        Cursor.CODEC.listOf()
            .encodeStart(class_2509.field_11560, this.cursors)
            .resultOrPartial(WildBackport.LOGGER::error)
            .ifPresent(value -> { tag.method_10566("cursors", value); });
    }

    public void spread(class_2338 pos, int charge) {
        while (charge > 0) {
            int spread = Math.min(charge, 1000);
            this.addCursor(new Cursor(pos, spread));
            charge -= spread;
        }
    }

    private void addCursor(Cursor cursor) {
        if (this.cursors.size() < 32) {
            this.cursors.add(cursor);
        }
    }

    public void
    tick(class_1936 level, class_2338 pos, Random random, boolean shouldConvert) {
        class_1937 instance = level instanceof class_1937 side ? side : null;
        if (!this.cursors.isEmpty()) {
            List<Cursor> cursors = new ArrayList<>();
            Map<class_2338, Cursor> cursorPositions = new HashMap<>();
            Object2IntMap<class_2338> positions = new Object2IntOpenHashMap<>();

            for (Cursor cursor : this.cursors) {
                cursor.spread(level, pos, random, this, shouldConvert);
                if (cursor.charge <= 0) {
                    applySculkCharge(instance, cursor.getPos(), 0);
                } else {
                    class_2338 position = cursor.getPos();
                    positions.computeInt(
                        position,
                        (blockPos,
                         charge) -> (charge == null ? 0 : charge) + cursor.charge
                    );
                    Cursor target = cursorPositions.get(position);
                    if (target == null) {
                        cursorPositions.put(position, cursor);
                        cursors.add(cursor);
                    } else if (!this.isWorldGen() && cursor.charge + target.charge <= 1000) {
                        target.merge(cursor);
                    } else {
                        cursors.add(cursor);
                        if (cursor.charge < target.charge) {
                            cursorPositions.put(position, cursor);
                        }
                    }
                }
            }

            for (Object2IntMap.Entry<class_2338> entry : positions.object2IntEntrySet()) {
                class_2338 position = entry.getKey();
                int exp = entry.getIntValue();
                Cursor cursor = cursorPositions.get(position);
                Set<class_2350> directions = cursor == null ? null : cursor.getFacings();
                if (exp > 0 && directions != null) {
                    int charge = (int) (Math.log1p(exp) / (double) 2.3F) + 1;
                    int data
                        = (charge << 6) + SculkVeinBlock.directionsToFlag(directions);
                    applySculkCharge(instance, cursor.getPos(), data);
                }
            }

            this.cursors = cursors;
        }
    }

    public static class Cursor {
        private static final ObjectArrayList<class_2382> OFFSETS
            = class_156.method_654(new ObjectArrayList<>(18), positions -> {
                  class_2338
                      .method_20437(
                          new class_2338(-1, -1, -1), new class_2338(1, 1, 1)
                      )
                      .filter(pos -> {
                          return (pos.method_10263() == 0 || pos.method_10264() == 0 || pos.method_10260() == 0)
                              && pos != class_2338.field_10980;
                      })
                      .map(class_2338::method_10062)
                      .forEach(positions::add);
              });
        private class_2338 pos;
        private int charge;
        private int updateDelay;
        private int decayDelay;
        @Nullable
        private Set<class_2350> facings;
        private static final Codec<Set<class_2350>> DIRECTION_SET
            = class_2350.field_29502.listOf().xmap(
                directions
                -> Sets.newEnumSet(directions, class_2350.class),
                Lists::newArrayList
            );
        public static final Codec<Cursor> CODEC = RecordCodecBuilder.create(instance -> {
            return instance
                .group(
                    class_2338.field_25064.fieldOf("pos").forGetter(Cursor::getPos),
                    Codec.intRange(0, 1000).fieldOf("charge").orElse(0).forGetter(
                        Cursor::getCharge
                    ),
                    Codec.intRange(0, 1)
                        .fieldOf("decay_delay")
                        .orElse(1)
                        .forGetter(Cursor::getDecayDelay),
                    Codec.intRange(0, Integer.MAX_VALUE)
                        .fieldOf("update_delay")
                        .orElse(0)
                        .forGetter(cursor -> { return cursor.updateDelay; }),
                    DIRECTION_SET.optionalFieldOf("facings").forGetter(cursor -> {
                        return Optional.ofNullable(cursor.getFacings());
                    })
                )
                .apply(instance, Cursor::new);
        });

        private Cursor(
            class_2338 pos,
            int charge,
            int decayDelay,
            int updateDelay,
            Optional<Set<class_2350>> facings
        ) {
            this.pos = pos;
            this.charge = charge;
            this.decayDelay = decayDelay;
            this.updateDelay = updateDelay;
            this.facings = facings.orElse(null);
        }

        public Cursor(class_2338 pos, int charge) {
            this(pos, charge, 1, 0, Optional.empty());
        }

        public class_2338 getPos() {
            return this.pos;
        }

        public int getCharge() {
            return this.charge;
        }

        public int getDecayDelay() {
            return this.decayDelay;
        }

        @Nullable
        public Set<class_2350> getFacings() {
            return this.facings;
        }

        private boolean canSpread(class_1936 level, class_2338 pos, boolean isWorldGen) {
            if (this.charge <= 0) {
                return false;
            } else if (isWorldGen) {
                return true;
            } else if (level instanceof class_3218 server) {
                return server.method_39425(class_1923.method_37232(pos));
            } else {
                return false;
            }
        }

        public void spread(
            class_1936 level,
            class_2338 pos,
            Random random,
            SculkSpreadManager spreadManager,
            boolean shouldConvert
        ) {
            if (this.canSpread(level, pos, spreadManager.isWorldGen)) {
                if (this.updateDelay > 0) {
                    --this.updateDelay;
                } else {
                    class_2680 state = level.method_8320(this.pos);
                    SculkSpreadable spreadable = getSpreadable(state);
                    if (shouldConvert
                        && spreadable.spread(
                            level,
                            this.pos,
                            state,
                            this.facings,
                            spreadManager.isWorldGen()
                        )) {
                        if (spreadable.shouldConvertToSpreadable()) {
                            state = level.method_8320(this.pos);
                            spreadable = getSpreadable(state);
                        }

                        level.method_8396(
                            null,
                            this.pos,
                            WBSoundEvents.BLOCK_SCULK_BREAK,
                            class_3419.field_15245,
                            1.0F,
                            1.0F
                        );
                    }

                    this.charge = spreadable.spread(
                        this, level, pos, random, spreadManager, shouldConvert
                    );
                    if (this.charge <= 0) {
                        spreadable.spreadAtSamePosition(level, state, this.pos, random);
                    } else {
                        class_2338 target = getSpreadPos(level, this.pos, random);
                        if (target != null) {
                            spreadable.spreadAtSamePosition(
                                level, state, this.pos, random
                            );
                            this.pos = target.method_10062();
                            if (spreadManager.isWorldGen()
                                && !this.pos.method_19771(
                                    new class_2382(pos.method_10263(), this.pos.method_10264(), pos.method_10260()),
                                    15.0D
                                )) {
                                this.charge = 0;
                                return;
                            }

                            state = level.method_8320(target);
                        }

                        if (state.method_26204() instanceof SculkSpreadable) {
                            this.facings = SculkVeinBlock.collectDirections(state);
                        }

                        this.decayDelay = spreadable.getDecay(this.decayDelay);
                        this.updateDelay = spreadable.getUpdate();
                    }
                }
            }
        }

        void merge(Cursor cursor) {
            this.charge += cursor.charge;
            cursor.charge = 0;
            this.updateDelay = Math.min(this.updateDelay, cursor.updateDelay);
        }

        private static SculkSpreadable getSpreadable(class_2680 state) {
            return state.method_26204() instanceof SculkSpreadable spreadable
                ? spreadable
                : SculkSpreadable.DEFAULT;
        }

        private static List<class_2382> shuffleOffsets(Random random) {
            return ModUtils.copyShuffled(OFFSETS, random);
        }

        @Nullable
        private static class_2338
        getSpreadPos(class_1936 level, class_2338 pos, Random random) {
            class_2338.class_2339 target = pos.method_25503();
            class_2338.class_2339 source = pos.method_25503();

            for (class_2382 offset : shuffleOffsets(random)) {
                source.method_35831(pos, offset);
                class_2680 state = level.method_8320(source);
                if (state.method_26204() instanceof SculkSpreadable
                    && canSpread(level, pos, source)) {
                    target.method_10101(source);
                    if (SculkVeinBlock.veinCoversSculkReplaceable(level, state, source)) {
                        break;
                    }
                }
            }

            return target.equals(pos) ? null : target;
        }

        private static boolean
        canSpread(class_1936 level, class_2338 source, class_2338 target) {
            if (source.method_19455(target) == 1) {
                return true;
            } else {
                class_2338 pos = target.method_10059(target);
                class_2350 xAxis = class_2350.method_10169(
                    class_2350.class_2351.field_11048,
                    pos.method_10263() < 0 ? class_2350.class_2352.field_11060
                                   : class_2350.class_2352.field_11056
                );
                class_2350 yAxis = class_2350.method_10169(
                    class_2350.class_2351.field_11052,
                    pos.method_10264() < 0 ? class_2350.class_2352.field_11060
                                   : class_2350.class_2352.field_11056
                );
                class_2350 zAxis = class_2350.method_10169(
                    class_2350.class_2351.field_11051,
                    pos.method_10260() < 0 ? class_2350.class_2352.field_11060
                                   : class_2350.class_2352.field_11056
                );
                if (pos.method_10263() == 0) {
                    return canSpread(level, source, yAxis)
                        || canSpread(level, source, zAxis);
                } else if (pos.method_10264() == 0) {
                    return canSpread(level, source, xAxis)
                        || canSpread(level, source, zAxis);
                } else {
                    return canSpread(level, source, xAxis)
                        || canSpread(level, source, yAxis);
                }
            }
        }

        private static boolean
        canSpread(class_1936 level, class_2338 pos, class_2350 direction) {
            class_2338 facing = pos.method_10093(direction);
            return !level.method_8320(facing).method_26206(
                level, facing, direction.method_10153()
            );
        }
    }

    public static void applySculkCharge(class_1937 level, class_2338 pos, int data) {
        if (level == null)
            return;

        Random random = level.method_8409();
        class_638 client
            = level.method_8608() && level instanceof class_638 side ? side : null;
        class_3218 server = level instanceof class_3218 side ? side : null;

        int charge = data >> 6;
        if (charge > 0) {
            if (random.nextFloat() < (float) charge * 0.2F) {
                float volume = 0.15F
                    + 0.05F * (float) charge * (float) charge * random.nextFloat();
                float pitch = 0.4F * (float) charge - 0.2F * random.nextFloat();
                if (client != null)
                    client.method_2947(
                        pos,
                        WBSoundEvents.BLOCK_SCULK_CHARGE,
                        class_3419.field_15245,
                        volume,
                        pitch,
                        false
                    );
            }

            int facings = (byte) (data & 63);
            class_6019 spread = class_6019.method_35017(0, charge);
            Supplier<class_243> velocities = () -> {
                return new class_243(
                    class_3532.method_15366(random, -0.005D, 0.005D),
                    class_3532.method_15366(random, -0.005D, 0.005D),
                    class_3532.method_15366(random, -0.005D, 0.005D)
                );
            };

            if (facings == 0) {
                for (class_2350 direction : class_2350.values()) {
                    float roll = direction == class_2350.field_11033 ? (float) Math.PI : 0.0F;
                    double offset
                        = direction == class_2350.field_11036 || direction == class_2350.field_11033
                        ? 0.32D
                        : 0.57D;
                    ParticleUtils.spawnParticles(
                        level,
                        pos,
                        new SculkChargeParticleOptions(roll),
                        spread,
                        direction,
                        velocities,
                        offset
                    );
                }
            } else {
                for (class_2350 direction : DirectionUtils.unpack((byte) data)) {
                    float roll = direction == class_2350.field_11036 ? (float) Math.PI : 0.0F;
                    ParticleUtils.spawnParticles(
                        level,
                        pos,
                        new SculkChargeParticleOptions(roll),
                        spread,
                        direction,
                        velocities,
                        0.35D
                    );
                }
            }
        } else {
            if (client != null)
                client.method_2947(
                    pos,
                    WBSoundEvents.BLOCK_SCULK_CHARGE,
                    class_3419.field_15245,
                    1.0F,
                    1.0F,
                    false
                );
            boolean fullBlock
                = level.method_8320(pos).method_26234(level, pos);
            int tries = fullBlock ? 40 : 20;
            float spread = fullBlock ? 0.45F : 0.25F;

            for (int i = 0; i < tries; i++) {
                float x = 2.0F * random.nextFloat() - 1.0F;
                float y = 2.0F * random.nextFloat() - 1.0F;
                float z = 2.0F * random.nextFloat() - 1.0F;
                if (server != null)
                    server.method_14199(
                        WBParticleTypes.SCULK_CHARGE_POP.get(),
                        (double) pos.method_10263() + 0.5D + (double) (x * spread),
                        (double) pos.method_10264() + 0.5D + (double) (y * spread),
                        (double) pos.method_10260() + 0.5D + (double) (z * spread),
                        1,
                        x * 0.07F,
                        y * 0.07F,
                        z * 0.07F,
                        0.0D
                    );
            }
        }
    }
}