package com.cursedcauldron.wildbackport.common.entities.brain;

import java.util.List;
import net.minecraft.class_1309;
import net.minecraft.class_2338;
import net.minecraft.class_3218;
import net.minecraft.class_3532;
import net.minecraft.class_3902;
import net.minecraft.class_4095;
import net.minecraft.class_4097;
import net.minecraft.class_4099;
import net.minecraft.class_4101;
import net.minecraft.class_4110;
import net.minecraft.class_4112;
import net.minecraft.class_4118;
import net.minecraft.class_4119;
import net.minecraft.class_4125;
import net.minecraft.class_4140;
import net.minecraft.class_4141;
import net.minecraft.class_4148;
import net.minecraft.class_4149;
import net.minecraft.class_4168;
import net.minecraft.class_4816;
import net.minecraft.class_4818;
import net.minecraft.class_4822;
import net.minecraft.class_5134;
import com.cursedcauldron.wildbackport.common.entities.Warden;
import com.cursedcauldron.wildbackport.common.entities.brain.warden.Digging;
import com.cursedcauldron.wildbackport.common.entities.brain.warden.Dismount;
import com.cursedcauldron.wildbackport.common.entities.brain.warden.Emerging;
import com.cursedcauldron.wildbackport.common.entities.brain.warden.FindRoarTarget;
import com.cursedcauldron.wildbackport.common.entities.brain.warden.ForgetAttackTarget;
import com.cursedcauldron.wildbackport.common.entities.brain.warden.GoToTargetLocation;
import com.cursedcauldron.wildbackport.common.entities.brain.warden.Roar;
import com.cursedcauldron.wildbackport.common.entities.brain.warden.SetWardenLookTarget;
import com.cursedcauldron.wildbackport.common.entities.brain.warden.Sniffing;
import com.cursedcauldron.wildbackport.common.entities.brain.warden.SonicBoom;
import com.cursedcauldron.wildbackport.common.entities.brain.warden.TryToSniff;
import com.cursedcauldron.wildbackport.common.registry.entity.WBActivities;
import com.cursedcauldron.wildbackport.common.registry.entity.WBMemoryModules;
import com.cursedcauldron.wildbackport.common.registry.entity.WBSensorTypes;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.mojang.datafixers.util.Pair;
import com.mojang.serialization.Dynamic;

//<>

public class WardenBrain {
    private static final int DIGGING_DURATION = class_3532.method_15386(100.0F);
    public static final int EMERGE_DURATION = class_3532.method_15386(133.59999F);
    public static final int ROAR_DURATION = class_3532.method_15386(84.0F);
    private static final int SNIFFING_DURATION = class_3532.method_15386(83.2F);
    private static final List<class_4149<? extends class_4148<? super Warden>>> SENSOR_TYPES
        = List.of(class_4149.field_18467, WBSensorTypes.WARDEN_ENTITY_SENSOR.get());
    private static final List<class_4140<?>> MEMORY_TYPES = List.of(
        class_4140.field_18441,
        class_4140.field_18442,
        class_4140.field_18444,
        class_4140.field_22354,
        class_4140.field_25360,
        class_4140.field_18446,
        class_4140.field_18445,
        class_4140.field_19293,
        class_4140.field_18449,
        class_4140.field_22355,
        class_4140.field_22475,
        class_4140.field_30243,
        WBMemoryModules.ROAR_TARGET.get(),
        WBMemoryModules.DISTURBANCE_LOCATION.get(),
        WBMemoryModules.RECENT_PROJECTILE.get(),
        WBMemoryModules.IS_SNIFFING.get(),
        WBMemoryModules.IS_EMERGING.get(),
        WBMemoryModules.ROAR_SOUND_DELAY.get(),
        WBMemoryModules.DIG_COOLDOWN.get(),
        WBMemoryModules.ROAR_SOUND_COOLDOWN.get(),
        WBMemoryModules.SNIFF_COOLDOWN.get(),
        WBMemoryModules.TOUCH_COOLDOWN.get(),
        WBMemoryModules.VIBRATION_COOLDOWN.get(),
        WBMemoryModules.SONIC_BOOM_COOLDOWN.get(),
        WBMemoryModules.SONIC_BOOM_SOUND_COOLDOWN.get(),
        WBMemoryModules.SONIC_BOOM_SOUND_DELAY.get()
    );
    private static final class_4097<Warden> DIG_COOLDOWN_SETTER = new class_4097<>(
        ImmutableMap.of(WBMemoryModules.DIG_COOLDOWN.get(), class_4141.field_18458)
    ) {
        @Override
        protected void start(class_3218 level, Warden warden, long time) {
            resetDigCooldown(warden);
        }
    };

    public static void updateActivities(Warden warden) {
        warden.method_18868().method_24531(ImmutableList.of(
            WBActivities.EMERGE.get(),
            WBActivities.DIG.get(),
            WBActivities.ROAR.get(),
            class_4168.field_22396,
            WBActivities.INVESTIGATE.get(),
            WBActivities.SNIFF.get(),
            class_4168.field_18595
        ));
    }

    public static class_4095<?> makeBrain(Warden warden, Dynamic<?> dynamic) {
        class_4095.class_5303<Warden> provider = class_4095.method_28311(MEMORY_TYPES, SENSOR_TYPES);
        class_4095<Warden> brain = provider.method_28335(dynamic);
        initCoreActivity(brain);
        initEmergeActivity(brain);
        initDiggingActivity(brain);
        initIdleActivity(brain);
        initRoarActivity(brain);
        initFightActivity(warden, brain);
        initInvestigateActivity(brain);
        initSniffingActivity(brain);
        brain.method_18890(ImmutableSet.of(class_4168.field_18594));
        brain.method_18897(class_4168.field_18595);
        brain.method_24536();
        return brain;
    }

    private static void initCoreActivity(class_4095<Warden> brain) {
        brain.method_18882(
            class_4168.field_18594,
            0,
            ImmutableList.of(
                new class_4125(0.8F),
                new SetWardenLookTarget(),
                new class_4110(45, 90),
                new class_4112()
            )
        );
    }

    private static void initEmergeActivity(class_4095<Warden> brain) {
        brain.method_24527(
            WBActivities.EMERGE.get(),
            5,
            ImmutableList.of(new Emerging<>(EMERGE_DURATION)),
            WBMemoryModules.IS_EMERGING.get()
        );
    }

    private static void initDiggingActivity(class_4095<Warden> brain) {
        brain.method_24529(
            WBActivities.DIG.get(),
            ImmutableList.of(
                Pair.of(0, new Dismount()), Pair.of(1, new Digging<>(DIGGING_DURATION))
            ),
            ImmutableSet.of(
                Pair.of(WBMemoryModules.ROAR_TARGET.get(), class_4141.field_18457),
                Pair.of(WBMemoryModules.DIG_COOLDOWN.get(), class_4141.field_18457)
            )
        );
    }

    private static void initIdleActivity(class_4095<Warden> brain) {
        brain.method_18882(
            class_4168.field_18595,
            10,
            ImmutableList.of(
                new FindRoarTarget<>(Warden::getPrimeSuspect),
                new TryToSniff(),
                new class_4118<>(
                    ImmutableMap.of(
                        WBMemoryModules.IS_SNIFFING.get(), class_4141.field_18457
                    ),
                    ImmutableList.of(
                        Pair.of(new class_4818(0.5F), 2),
                        Pair.of(new class_4101(30, 60), 1)
                    )
                )
            )
        );
    }

    private static void initInvestigateActivity(class_4095<Warden> brain) {
        brain.method_24527(
            WBActivities.INVESTIGATE.get(),
            5,
            ImmutableList.of(
                new FindRoarTarget<>(Warden::getPrimeSuspect),
                new GoToTargetLocation<>(
                    WBMemoryModules.DISTURBANCE_LOCATION.get(), 2, 0.7F
                )
            ),
            WBMemoryModules.DISTURBANCE_LOCATION.get()
        );
    }

    private static void initSniffingActivity(class_4095<Warden> brain) {
        brain.method_24527(
            WBActivities.SNIFF.get(),
            5,
            ImmutableList.of(
                new FindRoarTarget<>(Warden::getPrimeSuspect),
                new Sniffing<>(SNIFFING_DURATION)
            ),
            WBMemoryModules.IS_SNIFFING.get()
        );
    }

    private static void initRoarActivity(class_4095<Warden> brain) {
        brain.method_24527(
            WBActivities.ROAR.get(),
            10,
            ImmutableList.of(new Roar()),
            WBMemoryModules.ROAR_TARGET.get()
        );
    }

    private static void initFightActivity(Warden warden, class_4095<Warden> brain) {
        brain.method_24527(
            class_4168.field_22396,
            10,
            ImmutableList.of(
                DIG_COOLDOWN_SETTER,
                new ForgetAttackTarget<>(
                    entity
                    -> !warden.getAngriness().isAngry() || !warden.isValidTarget(entity),
                    WardenBrain::removeDeadSuspect,
                    false
                ),
                new class_4119(
                    entity
                    -> isTargeting(warden, entity),
                    (float) warden.method_26825(class_5134.field_23717)
                ),
                new class_4822(1.2F),
                new SonicBoom(),
                new class_4816(18)
            ),
            class_4140.field_22355
        );
    }

    private static boolean isTargeting(Warden warden, class_1309 entity) {
        return warden.method_18868()
            .method_18904(class_4140.field_22355)
            .filter(target -> target == entity)
            .isPresent();
    }

    private static void removeDeadSuspect(Warden warden, class_1309 suspect) {
        if (!warden.isValidTarget(suspect))
            warden.removeSuspect(suspect);
        resetDigCooldown(warden);
    }

    public static void resetDigCooldown(class_1309 warden) {
        if (warden.method_18868().method_18896(WBMemoryModules.DIG_COOLDOWN.get())) {
            warden.method_18868().method_24525(
                WBMemoryModules.DIG_COOLDOWN.get(), class_3902.field_17274, 1200L
            );
        }
    }

    public static void lookAtDisturbance(Warden warden, class_2338 pos) {
        if (warden.field_6002.method_8621().method_11952(pos)
            && warden.getPrimeSuspect().isEmpty()
            && warden.method_18868().method_18904(class_4140.field_22355).isEmpty()) {
            resetDigCooldown(warden);
            warden.method_18868().method_24525(
                WBMemoryModules.SNIFF_COOLDOWN.get(), class_3902.field_17274, 100L
            );
            warden.method_18868().method_24525(
                class_4140.field_18446, new class_4099(pos), 100L
            );
            warden.method_18868().method_24525(
                WBMemoryModules.DISTURBANCE_LOCATION.get(), pos, 100L
            );
            warden.method_18868().method_18875(class_4140.field_18445);
        }
    }
}