package com.cursedcauldron.wildbackport.common.entities.brain.frog;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.UUID;
import net.minecraft.class_11;
import net.minecraft.class_1297;
import net.minecraft.class_1309;
import net.minecraft.class_1589;
import net.minecraft.class_1799;
import net.minecraft.class_1802;
import net.minecraft.class_3218;
import net.minecraft.class_3414;
import net.minecraft.class_3419;
import net.minecraft.class_4050;
import net.minecraft.class_4097;
import net.minecraft.class_4140;
import net.minecraft.class_4141;
import net.minecraft.class_4142;
import net.minecraft.class_4215;
import com.cursedcauldron.wildbackport.common.entities.Frog;
import com.cursedcauldron.wildbackport.common.entities.access.api.Poses;
import com.cursedcauldron.wildbackport.common.registry.WBBlocks;
import com.cursedcauldron.wildbackport.common.registry.entity.WBMemoryModules;
import com.google.common.collect.ImmutableMap;

//<>

public class FrogEat extends class_4097<Frog> {
    private int eatTick;
    private int moveToTargetTick;
    private final class_3414 tongueSound;
    private final class_3414 eatSound;
    private Phase phase = Phase.DONE;

    public FrogEat(class_3414 tongueSound, class_3414 eatSound) {
        super(
            ImmutableMap.of(
                class_4140.field_18445,
                class_4141.field_18457,
                class_4140.field_18446,
                class_4141.field_18458,
                class_4140.field_22355,
                class_4141.field_18456
            ),
            100
        );
        this.tongueSound = tongueSound;
        this.eatSound = eatSound;
    }

    @Override
    protected boolean checkExtraStartConditions(class_3218 level, Frog frog) {
        class_1309 target
            = frog.method_18868().method_18904(class_4140.field_22355).get();
        boolean flag = this.canMoveToTarget(frog, target);
        if (!flag) {
            frog.method_18868().method_18875(class_4140.field_22355);
            this.cantReachTarget(frog, target);
        }

        return flag && frog.method_18376() != Poses.CROAKING.get()
            && Frog.isValidFrogFood(target);
    }

    @Override
    protected boolean canStillUse(class_3218 level, Frog frog, long time) {
        return frog.method_18868().method_18896(class_4140.field_22355)
            && this.phase != Phase.DONE;
    }

    @Override
    protected void start(class_3218 level, Frog frog, long time) {
        class_1309 entity
            = frog.method_18868().method_18904(class_4140.field_22355).get();
        class_4215.method_19554(frog, entity);
        frog.setFrogTarget(entity);
        frog.method_18868().method_18878(
            class_4140.field_18445, new class_4142(entity.method_19538(), 2.0F, 0)
        );
        this.moveToTargetTick = 10;
        this.phase = Phase.MOVE_TO_TARGET;
    }

    @Override
    protected void stop(class_3218 level, Frog frog, long time) {
        frog.method_18868().method_18875(class_4140.field_22355);
        frog.clearFrogTarget();
        frog.method_18380(class_4050.field_18076);
    }

    private void eat(class_3218 level, Frog frog) {
        level.method_8449(null, frog, this.eatSound, class_3419.field_15254, 2.0F, 1.0F);
        Optional<class_1297> target = frog.getFrogTarget();
        if (target.isPresent()) {
            class_1297 entity = target.get();
            if (entity.method_5805()) {
                frog.method_6121(entity);
                if (!entity.method_5805()) {
                    entity.method_5775(dropStack(frog, entity));
                    entity.method_5650(class_1297.class_5529.field_26998);
                }
            }
        }
    }

    //Change this to loot table
    private static class_1799 dropStack(Frog frog, class_1297 entity) {
        if (entity instanceof class_1589) {
            return new class_1799(switch (frog.getVariant()) {
                case TEMPERATE -> WBBlocks.OCHRE_FROGLIGHT.get().asItem();
                case WARM -> WBBlocks.PEARLESCENT_FROGLIGHT.get().asItem();
                case COLD -> WBBlocks.VERDANT_FROGLIGHT.get().asItem();
            });
        } else {
            return new class_1799(class_1802.field_8777);
        }
    }

    @Override
    protected void tick(class_3218 level, Frog frog, long time) {
        class_1309 target = frog.method_18868().method_18904(class_4140.field_22355).get();
        frog.setFrogTarget(target);
        switch (this.phase) {
            case MOVE_TO_TARGET:
                    if (target.method_5739(frog) < 1.75F) {
                        level.method_8449(
                            null, frog, this.tongueSound, class_3419.field_15254, 2.0F, 1.0F
                        );
                        frog.method_18380(Poses.USING_TONGUE.get());
                        target.method_18799(target.method_19538()
                                                    .method_1035(frog.method_19538())
                                                    .method_1029()
                                                    .method_1021(0.75D));
                        this.eatTick = 0;
                        this.phase = Phase.CATCH_ANIMATION;
                    } else if (this.moveToTargetTick <= 0) {
                        frog.method_18868().method_18878(
                            class_4140.field_18445,
                            new class_4142(target.method_19538(), 2.0F, 0)
                        );
                        this.moveToTargetTick = 10;
                    } else {
                        --this.moveToTargetTick;
                    }
                    break;
                case CATCH_ANIMATION:
                    if (this.eatTick++ >= 6) {
                        this.phase = Phase.EAT_ANIMATION;
                        this.eat(level, frog);
                    }
                    break;
                case EAT_ANIMATION:
                    if (this.eatTick >= 10) {
                        this.phase = Phase.DONE;
                    } else {
                        ++this.eatTick;
                    }
                case DONE:
        }
        }

        private boolean canMoveToTarget(Frog frog, class_1309 entity) {
            class_11 path = frog.method_5942().method_6349(entity, 0);
            return path != null && path.method_21656() < 1.75F;
        }

        private void cantReachTarget(Frog frog, class_1309 entity) {
            List<UUID> targets
                = frog.method_18868()
                      .method_18904(WBMemoryModules.UNREACHABLE_TONGUE_TARGETS.get())
                      .orElseGet(ArrayList::new);
            boolean notTargeting = !targets.contains(entity.method_5667());
            if (targets.size() == 5 && notTargeting)
                targets.remove(0);

            if (notTargeting)
                targets.add(entity.method_5667());

            frog.method_18868().method_24525(
                WBMemoryModules.UNREACHABLE_TONGUE_TARGETS.get(), targets, 100L
            );
        }

        enum Phase { MOVE_TO_TARGET, CATCH_ANIMATION, EAT_ANIMATION, DONE }
    }