package com.cursedcauldron.wildbackport.common.entities.brain.frog;

import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import net.minecraft.class_11;
import net.minecraft.class_1308;
import net.minecraft.class_14;
import net.minecraft.class_1408;
import net.minecraft.class_2246;
import net.minecraft.class_2338;
import net.minecraft.class_238;
import net.minecraft.class_243;
import net.minecraft.class_2680;
import net.minecraft.class_3218;
import net.minecraft.class_3414;
import net.minecraft.class_3419;
import net.minecraft.class_3532;
import net.minecraft.class_4048;
import net.minecraft.class_4050;
import net.minecraft.class_4097;
import net.minecraft.class_4099;
import net.minecraft.class_4140;
import net.minecraft.class_4141;
import net.minecraft.class_6008;
import net.minecraft.class_6011;
import net.minecraft.class_6019;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import org.jetbrains.annotations.Nullable;

public class FrogJumpToRandomPos<E extends class_1308> extends class_4097<E> {
    private static final List<Integer> ANGLES = Lists.newArrayList(65, 70, 75, 80);
    private final class_6019 cooldown;
    protected final int yRange;
    protected final int xzRange;
    protected final float maxRange;
    protected List<Target> targets = Lists.newArrayList();
    protected Optional<class_243> lastPos = Optional.empty();
    @Nullable
    protected class_243 lastTarget;
    protected int tries;
    protected long targetTime;
    private final Function<E, class_3414> landingSound;
    private final Predicate<class_2680> landingBlocks;

    public FrogJumpToRandomPos(
        class_6019 cooldown,
        int yRange,
        int xzRange,
        float range,
        Function<E, class_3414> landingSound,
        Predicate<class_2680> landingBlocks
    ) {
        super(
            ImmutableMap.of(
                class_4140.field_18446,
                class_4141.field_18458,
                class_4140.field_30244,
                class_4141.field_18457,
                class_4140.field_30245,
                class_4141.field_18457
            ),
            200
        );
        this.cooldown = cooldown;
        this.yRange = yRange;
        this.xzRange = xzRange;
        this.maxRange = range;
        this.landingSound = landingSound;
        this.landingBlocks = landingBlocks;
    }

    @Override
    protected boolean checkExtraStartConditions(class_3218 level, class_1308 entity) {
        boolean canUse = entity.method_24828() && !entity.method_5799() && !entity.method_5771()
            && !level.method_8320(entity.method_24515()).method_27852(class_2246.field_21211);
        if (!canUse)
            entity.method_18868().method_18878(
                class_4140.field_30244,
                this.cooldown.method_35008(level.field_9229) / 2
            );
        return canUse;
    }

    @Override
    protected boolean canStillUse(class_3218 level, class_1308 entity, long time) {
        boolean canUse = this.lastPos.isPresent()
            && this.lastPos.get().equals(entity.method_19538()) && this.tries > 0
            && !entity.method_5816()
            && (this.lastTarget != null || !this.targets.isEmpty());
        if (!canUse
            && entity.method_18868()
                   .method_18904(class_4140.field_30245)
                   .isEmpty()) {
            entity.method_18868().method_18878(
                class_4140.field_30244,
                this.cooldown.method_35008(level.field_9229) / 2
            );
            entity.method_18868().method_18875(class_4140.field_18446);
        }

        return canUse;
    }

    @Override
    protected void start(class_3218 level, E entity, long time) {
        this.lastTarget = null;
        this.tries = 20;
        this.lastPos = Optional.of(entity.method_19538());
        class_2338 pos = entity.method_24515();
        int x = pos.method_10263();
        int y = pos.method_10264();
        int z = pos.method_10260();
        this.targets
            = class_2338
                  .method_17962(
                      x - this.xzRange,
                      y - this.yRange,
                      z - this.xzRange,
                      x + this.xzRange,
                      y + this.yRange,
                      z + this.xzRange
                  )
                  .filter(position -> !position.equals(pos))
                  .map(
                      position
                      -> new Target(position.method_10062(), class_3532.method_15384(pos.method_10262(position)))
                  )
                  .collect(Collectors.toCollection(Lists::newArrayList));
    }

    @Override
    protected void tick(class_3218 level, E entity, long time) {
        if (this.lastTarget != null) {
            if (time - this.targetTime >= 40L) {
                entity.method_36456(entity.field_6283);
                entity.method_35054(true);
                double length = this.lastTarget.method_1033();
                double height = length + entity.method_37416();
                entity.method_18799(this.lastTarget.method_1021(height / length));
                entity.method_18868().method_18878(class_4140.field_30245, true);
                level.method_8449(
                    null,
                    entity,
                    this.landingSound.apply(entity),
                    class_3419.field_15254,
                    1.0F,
                    1.0F
                );
            }
        } else {
            --this.tries;
            this.pickTarget(level, entity, time);
        }
    }

    protected void pickTarget(class_3218 level, E entity, long time) {
        while (true) {
            if (!this.targets.isEmpty()) {
                Optional<Target> jumpTarget = this.jumpTarget(level);
                if (jumpTarget.isEmpty())
                    continue;

                Target target = jumpTarget.get();
                class_2338 pos = target.getPos();
                if (!canLandOn(level, entity, pos))
                    continue;

                class_243 center = class_243.method_24953(pos);
                class_243 lastTarget = this.getRammingVelocity(entity, center);
                if (lastTarget == null)
                    continue;

                entity.method_18868().method_18878(
                    class_4140.field_18446, new class_4099(pos)
                );
                class_1408 navigation = entity.method_5942();
                class_11 path = navigation.method_35141(pos, 0, 8);
                if (path != null && path.method_21655())
                    continue;

                this.lastTarget = lastTarget;
                this.targetTime = time;
                return;
            }

            return;
        }
    }

    protected Optional<Target> jumpTarget(class_3218 level) {
        Optional<Target> target
            = class_6011.method_34986(level.field_9229, this.targets);
        target.ifPresent(this.targets::remove);
        return target;
    }

    protected boolean canLandOn(class_3218 level, E entity, class_2338 pos) {
        class_2338 position = entity.method_24515();
        int x = position.method_10263();
        int z = position.method_10260();
        if (x == pos.method_10263() && z == pos.method_10260()) {
            return false;
        } else if (!entity.method_5942().method_6333(pos) && !this.landingBlocks.test(level.method_8320(pos.method_10074()))) {
            return false;
        } else {
            return entity.method_5944(class_14.method_23476(
                       entity.field_6002, pos.method_25503()
                   ))
                == 0.0F;
        }
    }

    @Nullable
    protected class_243 getRammingVelocity(class_1308 entity, class_243 pos) {
        List<Integer> angles = Lists.newArrayList(ANGLES);
        Collections.shuffle(angles);

        for (int angle : angles) {
            class_243 velocity = this.getRammingVelocity(entity, pos, angle);
            if (velocity != null)
                return velocity;
        }

        return null;
    }

    @Nullable
    private class_243 getRammingVelocity(class_1308 entity, class_243 pos, int angle) {
        class_243 position = entity.method_19538();
        class_243 scale = new class_243(pos.field_1352 - position.field_1352, 0.0, pos.field_1350 - position.field_1350)
                         .method_1029()
                         .method_1021(0.5D);
        pos = pos.method_1020(scale);
        class_243 distance = pos.method_1020(position);
        float maxAngle = (float) angle * (float) Math.PI / 180.0F;
        double xzRange = Math.atan2(distance.field_1350, distance.field_1352);
        double yRange = distance.method_1023(0.0D, distance.field_1351, 0.0D).method_1027();
        double yRadius = Math.sqrt(yRange);
        double i = Math.sin(2.0F * maxAngle);
        double k = Math.pow(Math.cos(maxAngle), 2.0D);
        double yMax = Math.sin(maxAngle);
        double xzMax = Math.cos(maxAngle);
        double zOffset = Math.sin(xzRange);
        double xOffset = Math.cos(xzRange);
        double jumpHeight = yRange * 0.08 / (yRadius * i - 2.0 * distance.field_1351 * k);
        if (jumpHeight < 0.0) {
            return null;
        } else {
            double range = Math.sqrt(jumpHeight);
            if (range > (double) this.maxRange) {
                return null;
            } else {
                double xzDistance = range * xzMax;
                double yDistance = range * yMax;
                int radius = class_3532.method_15384(yRadius / xzDistance) * 2;
                double index = 0.0;
                class_243 source = null;

                for (int j = 0; j < radius - 1; ++j) {
                    index += yRadius / (double) radius;
                    double x = index * xOffset;
                    double y = yMax / xzMax * index
                        - Math.pow(index, 2.0) * 0.08
                            / (2.0 * jumpHeight * Math.pow(xzMax, 2.0));
                    double z = index * zOffset;
                    class_243 target
                        = new class_243(position.field_1352 + x, position.field_1351 + y, position.field_1350 + z);
                    if (source != null && !this.canReach(entity, source, target)) {
                        return null;
                    }

                    source = target;
                }

                return new class_243(xzDistance * xOffset, yDistance, xzDistance * zOffset)
                    .method_1021(0.95F);
            }
        }
    }

    private boolean canReach(class_1308 entity, class_243 source, class_243 target) {
        class_4048 dimensions = entity.method_18377(class_4050.field_30095);
        class_243 distance = target.method_1020(source);
        double size = Math.min(dimensions.field_18067, dimensions.field_18068);
        int height = class_3532.method_15384(distance.method_1033() / size);
        class_243 normal = distance.method_1029();
        class_243 vector = source;
        for (int i = 0; i < height; ++i) {
            vector = i == height - 1 ? target
                                     : vector.method_1019(normal.method_1021(size * (double) 0.9F));
            class_238 box = dimensions.method_30757(vector);
            if (!entity.field_6002.method_8587(entity, box)) {
                return false;
            }
        }

        return true;
    }

    public static class Target extends class_6008.class_6009 {
        private final class_2338 pos;

        public Target(class_2338 pos, int weight) {
            super(weight);
            this.pos = pos;
        }

        public class_2338 getPos() {
            return this.pos;
        }
    }
}