package com.cursedcauldron.wildbackport.common.entities.warden;

import java.util.Optional;
import java.util.UUID;
import net.minecraft.class_1297;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_239;
import net.minecraft.class_243;
import net.minecraft.class_3218;
import net.minecraft.class_3222;
import net.minecraft.class_3481;
import net.minecraft.class_3532;
import net.minecraft.class_5698;
import net.minecraft.class_5699;
import net.minecraft.class_5702;
import net.minecraft.class_5712;
import net.minecraft.class_5714;
import net.minecraft.class_5716;
import net.minecraft.class_5720;
import net.minecraft.class_6862;
import Vibration;
import com.cursedcauldron.wildbackport.client.registry.WBCriteriaTriggers;
import com.cursedcauldron.wildbackport.common.utils.PositionUtils;
import com.mojang.serialization.Codec;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import org.jetbrains.annotations.Nullable;
import record;

//<>

public class VibrationHandler implements class_5714 {
    protected final class_5716 source;
    protected final int range;
    protected final VibrationConfig config;
    @Nullable
    protected Vibration event;
    protected float distance;
    protected int delay;

    public static Codec<VibrationHandler> codec(VibrationConfig config) {
        return RecordCodecBuilder.create(instance -> {
            return instance
                .group(
                    class_5716.field_28184.fieldOf("source").forGetter(listener -> {
                        return listener.source;
                    }),
                    class_5699.field_33441.fieldOf("range").forGetter(listener -> {
                        return listener.range;
                    }),
                    Vibration.CODEC.optionalFieldOf("event").forGetter(listener -> {
                        return Optional.ofNullable(listener.event);
                    }),
                    Codec.floatRange(0.0F, Float.MAX_VALUE)
                        .fieldOf("event_distance")
                        .orElse(0.0F)
                        .forGetter(listener -> { return listener.distance; }),
                    class_5699.field_33441.fieldOf("event_delay")
                        .orElse(0)
                        .forGetter(listener -> { return listener.delay; })
                )
                .apply(instance, (source, range, event, distance, delay) -> {
                    return new VibrationHandler(
                        source, range, config, event.orElse(null), distance, delay
                    );
                });
        });
    }

    public VibrationHandler(
        class_5716 source,
        int range,
        VibrationConfig config,
        @Nullable Vibration event,
        float distance,
        int delay
    ) {
        this.source = source;
        this.range = range;
        this.config = config;
        this.event = event;
        this.distance = distance;
        this.delay = delay;
    }

    public VibrationHandler(class_5716 source, int range, VibrationConfig config) {
        this(source, range, config, null, 0.0F, 0);
    }

    public void tick(class_1937 level) {
        if (level instanceof class_3218 server) {
            if (this.event != null) {
                --this.delay;
                if (this.delay <= 0) {
                    this.delay = 0;
                    this.config.onSignalReceive(
                        server,
                        this,
                        new class_2338(this.event.pos),
                        this.event.event,
                        this.event.getEntity(server).orElse(null),
                        this.event.getProjectileOwner(server).orElse(null),
                        this.distance
                    );
                    this.event = null;
                }
            }
        }
    }

    @Override
    public class_5716 method_32946() {
        return this.source;
    }

    @Override
    public int method_32948() {
        return this.range;
    }

    @Override
    public boolean
    method_32947(class_1937 level, class_5712 event, @Nullable class_1297 entity, class_2338 pos) {
        if (this.event != null) {
            return false;
        } else {
            Optional<class_2338> optional = this.source.method_32956(level);
            if (!this.config.isValidVibration(event, entity)) {
                return false;
            } else {
                class_243 source = PositionUtils.toVec(pos);
                class_243 target = PositionUtils.toVec(optional.get());
                if (!this.config.shouldListen(
                        (class_3218) level, this, new class_2338(source), event, entity
                    )) {
                    return false;
                } else if (isOccluded(level, source, target)) {
                    return false;
                } else {
                    this.scheduleSignal(level, event, entity, source, target);
                    return true;
                }
            }
        }
    }

    private void scheduleSignal(
        class_1937 level, class_5712 event, @Nullable class_1297 entity, class_243 source, class_243 target
    ) {
        this.distance = (float) source.method_1022(target);
        this.event = new Vibration(event, this.distance, source, entity);
        this.delay = class_3532.method_15375(this.distance);
        ((class_3218) level)
            .method_32817(new class_5720(
                PositionUtils.toBlockPos(source), this.source, this.delay
            ));
        this.config.onSignalSchedule();
    }

    private static boolean isOccluded(class_1937 level, class_243 source, class_243 target) {
        class_243 sourceVec = new class_243(
            (double) class_3532.method_15357(source.field_1352) + 0.5D,
            (double) class_3532.method_15357(source.field_1351) + 0.5D,
            (double) class_3532.method_15357(source.field_1350) + 0.5D
        );
        class_243 targetVec = new class_243(
            (double) class_3532.method_15357(target.field_1352) + 0.5D,
            (double) class_3532.method_15357(target.field_1351) + 0.5D,
            (double) class_3532.method_15357(target.field_1350) + 0.5D
        );

        for (class_2350 direction : class_2350.values()) {
            class_243 offsetVec = PositionUtils.relative(sourceVec, direction, 1.0E-5F);
            if (level
                    .method_32880(new class_5702(
                        offsetVec,
                        targetVec,
                        state -> {
                            return state.method_26164(class_3481.field_28088);
                        }
                    ))
                    .method_17783()
                != class_239.class_240.field_1332) {
                return false;
            }
        }

        return true;
    }

    public record Vibration(
        class_5712 event,
        float distance,
        class_243 pos,
        @Nullable UUID source,
        @Nullable UUID projectileOwner,
        @Nullable class_1297 entity
    ) {
        public static final Codec<Vibration> CODEC = RecordCodecBuilder.create(
            instance -> {
                return instance
                    .group(
                        Registry.GAME_EVENT.byNameCodec()
                            .fieldOf("game_event")
                            .forGetter(Vibration::event),
                        Codec.floatRange(0.0F, Float.MAX_VALUE)
                            .fieldOf("distance")
                            .forGetter(Vibration::distance),
                        PositionUtils.VEC_CODEC.fieldOf("pos").forGetter(Vibration::pos),
                        SerializableUUID.CODEC.optionalFieldOf("source").forGetter(
                            entity -> { return Optional.ofNullable(entity.source()); }
                        ),
                        SerializableUUID.CODEC.optionalFieldOf("projectile_owner")
                            .forGetter(entity -> {
                                return Optional.ofNullable(entity.projectileOwner());
                            })
                    )
                    .apply(instance, (event, distance, pos, source, projectileOwner) -> {
                        return new Vibration(
                            event,
                            distance,
                            pos,
                            source.orElse(null),
                            projectileOwner.orElse(null)
                        );
                    });
            }
        );

        public Vibration(
            GameEvent event,
            float distance,
            Vec3 pos,
            @Nullable UUID source,
            @Nullable UUID projectileOwner
        ) {
            this(event, distance, pos, source, projectileOwner, null);
        }

        public Vibration(
            GameEvent event, float distance, Vec3 pos, @Nullable Entity entity
        ) {
            this(
                event,
                distance,
                pos,
                entity == null ? null : entity.getUUID(),
                getProjectileOwner(entity),
                entity
            );
        }

        @Nullable
        private static UUID getProjectileOwner(@Nullable Entity entity) {
            if (entity instanceof Projectile projectile) {
                if (projectile.getOwner() != null) {
                    return projectile.getOwner().getUUID();
                }
            }

            return null;
        }

        public Optional<Entity> getEntity(ServerLevel level) {
            return Optional.ofNullable(this.entity).or(() -> {
                return Optional.ofNullable(this.source).map(level::getEntity);
            });
        }

        public Optional<Entity> getProjectileOwner(ServerLevel level) {
            return this.getEntity(level)
                .filter(entity -> { return entity instanceof Projectile; })
                .map(entity -> { return (Projectile) entity; })
                .map(Projectile::getOwner)
                .or(() -> {
                    return Optional.ofNullable(this.projectileOwner)
                        .map(level::getEntity);
                });
        }
    }

    public interface VibrationConfig {
        default class_6862<class_5712> getListenableEvents() {
            return class_5698.field_28090;
        }

        default boolean canTriggerAvoidVibration() {
            return false;
        }

        default boolean isValidVibration(class_5712 event, @Nullable class_1297 entity) {
            if (!event.method_40156(this.getListenableEvents())) {
                return false;
            } else {
                if (entity != null) {
                    if (entity.method_7325()) {
                        return false;
                    }

                    if (entity.method_21749()
                        && event.method_40156(class_5698.field_28091)) {
                        if (this.canTriggerAvoidVibration()
                            && entity instanceof class_3222 player) {
                            WBCriteriaTriggers.AVOID_VIBRATION.method_9027(player);
                        }

                        return false;
                    }

                    return !entity.method_33189();
                }

                return true;
            }
        }

        boolean shouldListen(
            class_3218 level,
            class_5714 listener,
            class_2338 pos,
            class_5712 event,
            @Nullable class_1297 entity
        );

        void onSignalReceive(
            class_3218 level,
            class_5714 listener,
            class_2338 pos,
            class_5712 event,
            @Nullable class_1297 entity,
            @Nullable class_1297 source,
            float distance
        );

        default void onSignalSchedule() {}
    }
}