package com.cursedcauldron.wildbackport.common.registry.worldgen;

import java.util.List;
import java.util.Optional;
import java.util.function.Supplier;
import net.minecraft.class_2246;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2378;
import net.minecraft.class_2382;
import net.minecraft.class_2975;
import net.minecraft.class_3031;
import net.minecraft.class_3141;
import net.minecraft.class_3226;
import net.minecraft.class_3612;
import net.minecraft.class_4651;
import net.minecraft.class_4659;
import net.minecraft.class_5204;
import net.minecraft.class_5450;
import net.minecraft.class_5458;
import net.minecraft.class_5780;
import net.minecraft.class_5928;
import net.minecraft.class_5929;
import net.minecraft.class_5934;
import net.minecraft.class_6016;
import net.minecraft.class_6019;
import net.minecraft.class_6646;
import net.minecraft.class_6658;
import net.minecraft.class_6732;
import net.minecraft.class_6792;
import net.minecraft.class_6793;
import net.minecraft.class_6796;
import net.minecraft.class_6817;
import net.minecraft.class_6880;
import net.minecraft.class_6885;
import com.cursedcauldron.wildbackport.common.blocks.MangrovePropaguleBlock;
import com.cursedcauldron.wildbackport.common.registry.WBBlocks;
import com.cursedcauldron.wildbackport.common.tag.WBBlockTags;
import com.cursedcauldron.wildbackport.common.utils.ModUtils;
import com.cursedcauldron.wildbackport.common.utils.SupplierHolder;
import com.cursedcauldron.wildbackport.common.worldgen.PredicatedStateProvider;
import com.cursedcauldron.wildbackport.common.worldgen.WorldGenerator;
import com.cursedcauldron.wildbackport.common.worldgen.decorator.AboveRootPlacement;
import com.cursedcauldron.wildbackport.common.worldgen.decorator.AttachedToLeavesDecorator;
import com.cursedcauldron.wildbackport.common.worldgen.decorator.MangroveRootPlacement;
import com.cursedcauldron.wildbackport.common.worldgen.decorator.WeightedLeaveVineDecorator;
import com.cursedcauldron.wildbackport.common.worldgen.features.GrassDiskConfiguration;
import com.cursedcauldron.wildbackport.common.worldgen.features.RootedTreeConfig;
import com.cursedcauldron.wildbackport.common.worldgen.features.SculkPatchConfiguration;
import com.cursedcauldron.wildbackport.common.worldgen.placers.MangroveRootPlacer;
import com.cursedcauldron.wildbackport.common.worldgen.placers.UpwardBranchingTrunk;
import com.cursedcauldron.wildbackport.core.api.CoreRegistry;

//<>

public class WBWorldGeneration {
    public static final CoreRegistry<class_2975<?, ?>> CONFIGURED_FEATURES
        = CoreRegistry.create(class_5458.field_25929, "minecraft");

    public static final CoreRegistry<class_6796> PLACED_FEATURES
        = CoreRegistry.create(class_5458.field_35761, "minecraft");

    public static void bootstrap() {
        WorldGenerator.setup();
    }

    // Mangrove Swamp
    public static final class_6880<class_2975<RootedTreeConfig, ?>> MANGROVE
        = SupplierHolder.createAndRegister(
            CONFIGURED_FEATURES,
            "mangrove",
            ()
                -> new class_2975<>(
                    WBFeatures.TREE.get(),
                    new RootedTreeConfig
                        .Builder(
                            class_4651.method_38432(WBBlocks.MANGROVE_LOG.get()),
                            new UpwardBranchingTrunk(
                                2,
                                1,
                                4,
                                class_6019.method_35017(1, 4),
                                0.5F,
                                class_6019.method_35017(0, 1),
                                class_2378.field_11146.method_40260(
                                    WBBlockTags.MANGROVE_LOGS_CAN_GROW_THROUGH
                                )
                            ),
                            class_4651.method_38432(WBBlocks.MANGROVE_LEAVES.get()),
                            new class_5928(
                                class_6016.method_34998(3),
                                class_6016.method_34998(0),
                                class_6016.method_34998(2),
                                70
                            ),
                            Optional.of(new MangroveRootPlacer(
                                class_6019.method_35017(1, 3),
                                class_4651.method_38432(WBBlocks.MANGROVE_ROOTS.get()),
                                Optional.of(new AboveRootPlacement(
                                    class_4651.method_38432(class_2246.field_28680), 0.5F
                                )),
                                new MangroveRootPlacement(
                                    class_2378.field_11146.method_40260(
                                        WBBlockTags.MANGROVE_ROOTS_CAN_GROW_THROUGH
                                    ),
                                    class_6885.method_40245(
                                        class_2248::method_40142,
                                        WBBlocks.MUD.get(),
                                        WBBlocks.MUDDY_MANGROVE_ROOTS.get()
                                    ),
                                    class_4651
                                        .method_38432(WBBlocks.MUDDY_MANGROVE_ROOTS.get()),
                                    8,
                                    15,
                                    0.2F
                                )
                            )),
                            new class_5204(2, 0, 2)
                        )
                        .decorators(List.of(
                            new WeightedLeaveVineDecorator(0.125F),
                            new AttachedToLeavesDecorator(
                                0.14F,
                                1,
                                0,
                                new class_5929(
                                    class_4651.method_38433(
                                        WBBlocks.MANGROVE_PROPAGULE.get()
                                            .method_9564()
                                            .method_11657(
                                                MangrovePropaguleBlock.HANGING, true
                                            )
                                    ),
                                    MangrovePropaguleBlock.AGE,
                                    class_6019.method_35017(0, 4)
                                ),
                                2,
                                List.of(class_2350.field_11033)
                            ),
                            new class_4659(0.01F)
                        ))
                        .ignoreVines()
                        .build()
                )
        );

    public static final class_6880<class_2975<RootedTreeConfig, ?>> TALL_MANGROVE
        = SupplierHolder.createAndRegister(
            CONFIGURED_FEATURES,
            "tall_mangrove",
            ()
                -> new class_2975<>(

                    WBFeatures.TREE.get(),
                    new RootedTreeConfig
                        .Builder(
                            class_4651.method_38432(WBBlocks.MANGROVE_LOG.get()),
                            new UpwardBranchingTrunk(
                                4,
                                1,
                                9,
                                class_6019.method_35017(1, 6),
                                0.5F,
                                class_6019.method_35017(0, 1),
                                class_2378.field_11146.method_40260(
                                    WBBlockTags.MANGROVE_LOGS_CAN_GROW_THROUGH
                                )
                            ),
                            class_4651.method_38432(WBBlocks.MANGROVE_LEAVES.get()),
                            new class_5928(
                                class_6016.method_34998(3),
                                class_6016.method_34998(0),
                                class_6016.method_34998(2),
                                70
                            ),
                            Optional.of(new MangroveRootPlacer(
                                class_6019.method_35017(3, 7),
                                class_4651.method_38432(WBBlocks.MANGROVE_ROOTS.get()),
                                Optional.of(new AboveRootPlacement(
                                    class_4651.method_38432(class_2246.field_28680), 0.5F
                                )),
                                new MangroveRootPlacement(
                                    class_2378.field_11146.method_40260(
                                        WBBlockTags.MANGROVE_ROOTS_CAN_GROW_THROUGH
                                    ),
                                    class_6885.method_40245(
                                        class_2248::method_40142,
                                        WBBlocks.MUD.get(),
                                        WBBlocks.MUDDY_MANGROVE_ROOTS.get()
                                    ),
                                    class_4651
                                        .method_38432(WBBlocks.MUDDY_MANGROVE_ROOTS.get()),
                                    8,
                                    15,
                                    0.2F
                                )
                            )),
                            new class_5204(3, 0, 2)
                        )
                        .decorators(List.of(
                            new WeightedLeaveVineDecorator(0.125F),
                            new AttachedToLeavesDecorator(
                                0.14F,
                                1,
                                0,
                                new class_5929(
                                    class_4651.method_38433(
                                        WBBlocks.MANGROVE_PROPAGULE.get()
                                            .method_9564()
                                            .method_11657(
                                                MangrovePropaguleBlock.HANGING, true
                                            )
                                    ),
                                    MangrovePropaguleBlock.AGE,
                                    class_6019.method_35017(0, 4)
                                ),
                                2,
                                List.of(class_2350.field_11033)
                            ),
                            new class_4659(0.01F)
                        ))
                        .ignoreVines()
                        .build()
                )
        );

    public static final class_6880<class_6796> MANGROVE_CHECKED
        = SupplierHolder.createAndRegister(
            PLACED_FEATURES,
            "mangrove_checked",
            ()
                -> new class_6796(
                    (class_6880) MANGROVE,
                    List.of(class_6817.method_40365(
                        WBBlocks.MANGROVE_PROPAGULE.get()
                    ))
                )
        );

    public static final class_6880<class_6796> TALL_MANGROVE_CHECKED
        = SupplierHolder.createAndRegister(
            PLACED_FEATURES,
            "tall_mangrove_checked",
            ()
                -> new class_6796(
                    (class_6880) TALL_MANGROVE,
                    List.of(class_6817.method_40365(
                        WBBlocks.MANGROVE_PROPAGULE.get()
                    ))
                )

        );

    public static final class_6880<class_2975<class_3141, ?>>
        MANGROVE_VEGETATION = SupplierHolder.createAndRegister(
            CONFIGURED_FEATURES,
            "mangrove_vegetation",
            ()
                -> new class_2975<>(
                    class_3031.field_13593,
                    new class_3141(
                        List.of(new class_3226(TALL_MANGROVE_CHECKED, 0.85F)),
                        MANGROVE_CHECKED
                    )
                )
        );

    public static final class_6880<class_6796> TREES_MANGROVE_PLACED
        = SupplierHolder.createAndRegister(
            PLACED_FEATURES,
            "trees_mangrove",
            ()
                -> new class_6796(
                    (class_6880) MANGROVE_VEGETATION,
                    List.of(
                        class_6793.method_39623(25),
                        class_5450.method_39639(),
                        class_5934.method_39662(5),
                        class_6817.field_36081,
                        class_6792.method_39614(),
                        class_6658.method_39618(class_6646.method_39009(
                            WBBlocks.MANGROVE_PROPAGULE.get().method_9564(),
                            class_2338.field_10980
                        ))

                    )
                )
        );

    public static final class_6880<class_2975<GrassDiskConfiguration, ?>> DISK_GRASS
        = SupplierHolder.createAndRegister(
            CONFIGURED_FEATURES,
            "disk_grass",
            ()
                -> new class_2975<>(
                    WBFeatures.DISK.get(),
                    new GrassDiskConfiguration(
                        new PredicatedStateProvider(
                            class_4651.method_38432(class_2246.field_10566),
                            List.of(new PredicatedStateProvider.Rule(
                                class_6646.method_38877(class_6646.method_38878(
                                    class_6646.method_39584(class_2350.field_11036.method_10163()),
                                    class_6646.method_38879(
                                        class_3612.field_15910, class_2350.field_11036.method_10163()
                                    )
                                )),
                                class_4651.method_38432(class_2246.field_10219)
                            ))
                        ),
                        class_6646.method_39910(
                            List.of(class_2246.field_10566, WBBlocks.MUD.get())
                        ),
                        class_6019.method_35017(2, 6),
                        2
                    )
                )
        );

    public static final class_6880<class_6796> DISK_GRASS_PLACED
        = SupplierHolder.createAndRegister(
            PLACED_FEATURES,
            "disk_grass",
            ()
                -> new class_6796(
                    (class_6880) DISK_GRASS,
                    List.of(
                        class_6793.method_39623(1),
                        class_5450.method_39639(),
                        class_6817.field_36079,
                        class_6732.method_39656(class_6016.method_34998(-1)),
                        class_6658.method_39618(
                            class_6646.method_38876(WBBlocks.MUD.get(), class_2382.field_11176)
                        ),
                        class_6792.method_39614()
                    )
                )

        );

    // Deep Dark
    public static final class_6880<class_2975<SculkPatchConfiguration, ?>>
        SCULK_PATCH_DEEP_DARK = SupplierHolder.createAndRegister(
            CONFIGURED_FEATURES,
            "sculk_patch_deep_dark",
            ()
                -> new class_2975<>(
                    WBFeatures.SCULK_PATCH.get(),
                    new SculkPatchConfiguration(10, 32, 64, 0, 1, class_6016.method_34998(0), 0.5F)
                )
        );

    public static final class_6880<class_6796> SCULK_PATCH_DEEP_DARK_PLACED
        = SupplierHolder.createAndRegister(
            PLACED_FEATURES,
            "sculk_patch_deep_dark",
            ()
                -> new class_6796(
                    (class_6880) SCULK_PATCH_DEEP_DARK,
                    List.of(
                        class_6793.method_39624(class_6016.method_34998(256)),
                        class_5450.method_39639(),
                        class_6817.field_36086,
                        class_6792.method_39614()
                    )
                )

        );

    public static final class_6880<class_2975<SculkPatchConfiguration, ?>>
        SCULK_PATCH_ANCIENT_CITY = SupplierHolder.createAndRegister(
            CONFIGURED_FEATURES,
            "sculk_patch_ancient_city",
            ()
                -> new class_2975<>(
                    WBFeatures.SCULK_PATCH.get(),
                    new SculkPatchConfiguration(
                        10, 32, 64, 0, 1, class_6019.method_35017(1, 3), 0.5F
                    )
                )
        );

    public static final class_6880<class_6796> SCULK_PATCH_ANCIENT_CITY_PLACED
        = SupplierHolder.createAndRegister(
            PLACED_FEATURES,
            "sculk_patch_ancient_city",
            () -> new class_6796((class_6880) SCULK_PATCH_ANCIENT_CITY, List.of())
        );

    public static final class_6880<class_2975<class_5780, ?>> SCULK_VEIN
        = SupplierHolder.createAndRegister(
            CONFIGURED_FEATURES,
            "sculk_vein",
            ()
                -> new class_2975<>(
                    WBFeatures.SCULK_GROWTH.get(),
                    new class_5780(
                        20,
                        true,
                        true,
                        true,
                        1.0F,
                        class_6885.method_40245(
                            class_2248::method_40142,
                            class_2246.field_10340,
                            class_2246.field_10115,
                            class_2246.field_10508,
                            class_2246.field_10474,
                            class_2246.field_28049,
                            class_2246.field_27114,
                            class_2246.field_27165,
                            class_2246.field_28888
                        )
                    )
                )
        );

    public static final class_6880<class_6796> SCULK_VEIN_PLACED
        = SupplierHolder.createAndRegister(
            PLACED_FEATURES,
            "sculk_vein",
            ()
                -> new class_6796(
                    (class_6880) SCULK_VEIN,
                    List.of(
                        class_6793.method_39624(class_6019.method_35017(204, 250)),
                        class_5450.method_39639(),
                        class_6817.field_36086,
                        class_6792.method_39614()
                    )
                )
        );
}
