package com.cursedcauldron.wildbackport.common.worldgen.decorator;

import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.function.BiConsumer;
import net.minecraft.class_156;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2680;
import net.minecraft.class_3031;
import net.minecraft.class_3746;
import net.minecraft.class_4651;
import net.minecraft.class_4662;
import net.minecraft.class_4663;
import com.cursedcauldron.wildbackport.common.registry.worldgen.WBTreeDecorators;
import com.cursedcauldron.wildbackport.common.utils.ModUtils;
import com.mojang.serialization.Codec;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import it.unimi.dsi.fastutil.objects.ObjectArrayList;

public class AttachedToLeavesDecorator extends class_4662 {
    public static final Codec<AttachedToLeavesDecorator> CODEC
        = RecordCodecBuilder.create(instance -> {
              return instance
                  .group(
                      Codec.floatRange(0.0F, 1.0F)
                          .fieldOf("probability")
                          .forGetter(decorator -> { return decorator.probability; }),
                      Codec.intRange(0, 16)
                          .fieldOf("exclusion_radius_xz")
                          .forGetter(decorator -> {
                              return decorator.exclusionRadiusXZ;
                          }),
                      Codec.intRange(0, 16)
                          .fieldOf("exclusion_radius_y")
                          .forGetter(decorator -> { return decorator.exclusionRadiusY; }),
                      class_4651.field_24937.fieldOf("block_provider")
                          .forGetter(decorator -> { return decorator.blockProvider; }),
                      Codec.intRange(1, 16)
                          .fieldOf("required_empty_blocks")
                          .forGetter(decorator -> {
                              return decorator.requiredEmptyBlocks;
                          }),
                      class_2350.field_29502.listOf()
                          .fieldOf("directions")
                          .forGetter(decorator -> { return decorator.directions; })
                  )
                  .apply(instance, AttachedToLeavesDecorator::new);
          });
    protected final float probability;
    protected final int exclusionRadiusXZ;
    protected final int exclusionRadiusY;
    protected final class_4651 blockProvider;
    protected final int requiredEmptyBlocks;
    protected final List<class_2350> directions;

    public AttachedToLeavesDecorator(
        float probability,
        int exclusionRadiusXZ,
        int exclusionRadiusY,
        class_4651 blockProvider,
        int requiredEmptyBlocks,
        List<class_2350> directions
    ) {
        this.probability = probability;
        this.exclusionRadiusXZ = exclusionRadiusXZ;
        this.exclusionRadiusY = exclusionRadiusY;
        this.blockProvider = blockProvider;
        this.requiredEmptyBlocks = requiredEmptyBlocks;
        this.directions = directions;
    }

    @Override
    public void method_23469(
        class_3746 level,
        BiConsumer<class_2338, class_2680> replacer,
        Random random,
        List<class_2338> trunkPositions,
        List<class_2338> foliagePositions
    ) {
        HashSet<class_2338> positions = new HashSet<>();
        for (class_2338 pos :
             ModUtils.copyShuffled(new ObjectArrayList<>(foliagePositions), random)) {
            class_2350 direction = class_156.method_32309(this.directions, random);
            class_2338 offset = pos.method_10093(direction);
            if (!positions.contains(offset) && random.nextFloat() < this.probability
                && this.meetsRequiredEmptyBlocks(level, pos, direction)) {
                class_2338 minRadius = offset.method_10069(
                    -this.exclusionRadiusXZ,
                    -this.exclusionRadiusY,
                    -this.exclusionRadiusXZ
                );
                class_2338 maxRadius = offset.method_10069(
                    this.exclusionRadiusXZ, this.exclusionRadiusY, this.exclusionRadiusXZ
                );
                for (class_2338 position : class_2338.method_10097(minRadius, maxRadius)) {
                    positions.add(position.method_10062());
                }

                replacer.accept(offset, this.blockProvider.method_23455(random, offset));
            }
        }
    }

    private boolean meetsRequiredEmptyBlocks(
        class_3746 level, class_2338 pos, class_2350 direction
    ) {
        for (int i = 1; i <= this.requiredEmptyBlocks; i++) {
            if (!class_3031.method_27370(level, pos.method_10079(direction, i)))
                return false;
        }

        return true;
    }

    @Override
    protected class_4663<?> method_28893() {
        return WBTreeDecorators.ATTACHED_TO_LEAVES.get();
    }
}