package com.cursedcauldron.wildbackport.common.worldgen.features;

import java.util.List;
import java.util.Optional;
import net.minecraft.class_2246;
import net.minecraft.class_4643;
import net.minecraft.class_4647;
import net.minecraft.class_4651;
import net.minecraft.class_4662;
import net.minecraft.class_5141;
import net.minecraft.class_5201;
import com.cursedcauldron.wildbackport.common.worldgen.placers.RootPlacer;
import com.google.common.collect.ImmutableList;
import com.mojang.serialization.Codec;
import com.mojang.serialization.codecs.RecordCodecBuilder;

//<>

public class RootedTreeConfig extends class_4643 {
    public static final Codec<RootedTreeConfig> CODEC = RecordCodecBuilder.create(
        instance -> {
            return instance
                .group(
                    class_4651.field_24937.fieldOf("trunk_provider")
                        .forGetter(config -> { return config.field_21288; }),
                    class_5141.field_24972.fieldOf("trunk_placer").forGetter(config -> {
                        return config.field_24136;
                    }),
                    class_4651.field_24937.fieldOf("foliage_provider")
                        .forGetter(config -> { return config.field_29280; }),
                    class_4647.field_24931.fieldOf("foliage_placer").forGetter(config -> {
                        return config.field_24135;
                    }),
                    RootPlacer.CODEC.optionalFieldOf("root_placer").forGetter(config -> {
                        return config.rootPlacer;
                    }),
                    class_4651.field_24937.fieldOf("dirt_provider")
                        .forGetter(config -> { return config.field_29279; }),
                    class_5201.field_24922.fieldOf("minimum_size").forGetter(config -> {
                        return config.field_24137;
                    }),
                    class_4662.field_24962.listOf()
                        .fieldOf("decorators")
                        .forGetter(config -> { return config.field_21290; }),
                    Codec.BOOL.fieldOf("ignore_vines").orElse(false).forGetter(config -> {
                        return config.field_24138;
                    }),
                    Codec.BOOL.fieldOf("force_dirt").orElse(false).forGetter(config -> {
                        return config.field_29281;
                    })
                )
                .apply(instance, RootedTreeConfig::new);
        }
    );
    public final Optional<RootPlacer> rootPlacer;

    protected RootedTreeConfig(
        class_4651 trunkProvider,
        class_5141 trunkPlacer,
        class_4651 foliageProvider,
        class_4647 foliagePlacer,
        Optional<RootPlacer> rootPlacer,
        class_4651 dirtProvider,
        class_5201 minimumSize,
        List<class_4662> decorators,
        boolean ignoreVines,
        boolean forceDirt
    ) {
        super(
            trunkProvider,
            trunkPlacer,
            foliageProvider,
            foliagePlacer,
            dirtProvider,
            minimumSize,
            decorators,
            ignoreVines,
            forceDirt
        );
        this.rootPlacer = rootPlacer;
    }

    public static class Builder {
        public final class_4651 trunkProvider;
        private final class_5141 trunkPlacer;
        public final class_4651 foliageProvider;
        private final class_4647 foliagePlacer;
        private final Optional<RootPlacer> rootPlacer;
        private class_4651 dirtProvider;
        private final class_5201 minimumSize;
        private List<class_4662> decorators = ImmutableList.of();
        private boolean ignoreVines;
        private boolean forceDirt;

        public Builder(
            class_4651 trunkProvider,
            class_5141 trunkPlacer,
            class_4651 foliageProvider,
            class_4647 foliagePlacer,
            Optional<RootPlacer> rootPlacer,
            class_5201 minimumSize
        ) {
            this.trunkProvider = trunkProvider;
            this.trunkPlacer = trunkPlacer;
            this.foliageProvider = foliageProvider;
            this.dirtProvider = class_4651.method_38432(class_2246.field_10566);
            this.foliagePlacer = foliagePlacer;
            this.rootPlacer = rootPlacer;
            this.minimumSize = minimumSize;
        }

        public Builder(
            class_4651 trunkProvider,
            class_5141 trunkPlacer,
            class_4651 foliageProvider,
            class_4647 foliagePlacer,
            class_5201 minimumSize
        ) {
            this(
                trunkProvider,
                trunkPlacer,
                foliageProvider,
                foliagePlacer,
                Optional.empty(),
                minimumSize
            );
        }

        public Builder dirtProvider(class_4651 dirtProvider) {
            this.dirtProvider = dirtProvider;
            return this;
        }

        public Builder decorators(List<class_4662> decorators) {
            this.decorators = decorators;
            return this;
        }

        public Builder ignoreVines() {
            this.ignoreVines = true;
            return this;
        }

        public Builder forceDirt() {
            this.forceDirt = true;
            return this;
        }

        public RootedTreeConfig build() {
            return new RootedTreeConfig(
                this.trunkProvider,
                this.trunkPlacer,
                this.foliageProvider,
                this.foliagePlacer,
                this.rootPlacer,
                this.dirtProvider,
                this.minimumSize,
                this.decorators,
                this.ignoreVines,
                this.forceDirt
            );
        }
    }
}