package com.cursedcauldron.wildbackport.common.worldgen.features;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.OptionalInt;
import java.util.Random;
import java.util.Set;
import java.util.function.BiConsumer;
import net.minecraft.class_1936;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2382;
import net.minecraft.class_244;
import net.minecraft.class_251;
import net.minecraft.class_2680;
import net.minecraft.class_2741;
import net.minecraft.class_2944;
import net.minecraft.class_3031;
import net.minecraft.class_3341;
import net.minecraft.class_3481;
import net.minecraft.class_3499;
import net.minecraft.class_3746;
import net.minecraft.class_4647;
import net.minecraft.class_5281;
import net.minecraft.class_5821;
import com.cursedcauldron.wildbackport.common.worldgen.placers.UpwardBranchingTrunk;
import com.cursedcauldron.wildbackport.core.mixin.access.TreeFeatureAccessor;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import com.mojang.serialization.Codec;

//<>

public class RootedTreeFeature extends class_3031<RootedTreeConfig> {
    public RootedTreeFeature(Codec<RootedTreeConfig> codec) {
        super(codec);
    }

    private boolean generate(
        class_5281 level,
        Random random,
        class_2338 pos,
        BiConsumer<class_2338, class_2680> rootPlacerReplacer,
        BiConsumer<class_2338, class_2680> trunkPlacerReplacer,
        BiConsumer<class_2338, class_2680> foliagePlacerReplacer,
        RootedTreeConfig config
    ) {
        int treeHeight = config.field_24136.method_26993(random);
        int foliageHeight
            = config.field_24135.method_26989(random, treeHeight, config);
        int trunkLength = treeHeight - foliageHeight;
        int foliageRadius = config.field_24135.method_23452(random, trunkLength);
        class_2338 origin
            = config.rootPlacer.map(rootPlacer -> rootPlacer.trunkOffset(pos, random))
                  .orElse(pos);
        int minHeight = Math.min(pos.method_10264(), origin.method_10264());
        int maxHeight = Math.max(pos.method_10264(), origin.method_10264()) + treeHeight + 1;
        if (minHeight >= level.method_31607() + 1
            && maxHeight <= level.method_31600()) {
            OptionalInt clippedHeight = config.field_24137.method_27377();
            int topPosition = this.getTopPosition(level, treeHeight, origin, config);
            if (topPosition >= treeHeight
                || clippedHeight.isPresent() && topPosition >= clippedHeight.getAsInt()) {
                if (config.rootPlacer.isPresent()
                    && !config.rootPlacer.get().generate(
                        level, rootPlacerReplacer, random, pos, origin, config
                    )) {
                    return false;
                } else {
                    List<class_4647.class_5208> foliage
                        = config.field_24136.method_26991(
                            level,
                            trunkPlacerReplacer,
                            random,
                            topPosition,
                            origin,
                            config
                        );
                    foliage.forEach(
                        node
                        -> config.field_24135.method_27385(
                            level,
                            foliagePlacerReplacer,
                            random,
                            config,
                            topPosition,
                            node,
                            foliageHeight,
                            foliageRadius
                        )
                    );
                    return true;
                }
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

    @SuppressWarnings("ConstantConditions")
    private int getTopPosition(
        class_3746 level, int height, class_2338 pos, RootedTreeConfig config
    ) {
        class_2338.class_2339 mutable = new class_2338.class_2339();
        for (int y = 0; y <= height + 1; ++y) {
            int size = config.field_24137.method_27378(height, y);

            for (int x = -size; x <= size; ++x) {
                for (int z = -size; z <= size; ++z) {
                    mutable.method_25504(pos, x, y, z);
                    boolean isValid = class_2944.method_27371(level, pos)
                        || level.method_16358(pos, state -> state.method_26164(class_3481.field_15475))
                        || (config.field_24136 instanceof UpwardBranchingTrunk trunk
                            && level.method_16358(
                                pos, state -> state.method_40143(trunk.canGrowThrough)
                            ));
                    if (!isValid
                        || (!config.field_24138
                            && TreeFeatureAccessor.isVine(level, mutable)))
                        return y - 2;
                }
            }
        }

        return height;
    }

    @Override
    public final boolean method_13151(class_5821<RootedTreeConfig> context) {
        class_5281 level = context.method_33652();
        Random random = context.method_33654();
        class_2338 pos = context.method_33655();
        RootedTreeConfig config = context.method_33656();
        HashSet<class_2338> rootPos = Sets.newHashSet();
        HashSet<class_2338> trunkPos = Sets.newHashSet();
        HashSet<class_2338> foliagePos = Sets.newHashSet();
        HashSet<class_2338> decoratorPos = Sets.newHashSet();
        BiConsumer<class_2338, class_2680> rootReplacer = (position, state) -> {
            rootPos.add(position.method_10062());
            level.method_8652(position, state, 19);
        };
        BiConsumer<class_2338, class_2680> trunkReplacer = (position, state) -> {
            trunkPos.add(position.method_10062());
            level.method_8652(position, state, 19);
        };
        BiConsumer<class_2338, class_2680> foliageReplacer = (position, state) -> {
            foliagePos.add(position.method_10062());
            level.method_8652(position, state, 19);
        };
        BiConsumer<class_2338, class_2680> decoratorReplacer = (position, state) -> {
            decoratorPos.add(position.method_10062());
            level.method_8652(position, state, 19);
        };
        boolean generate = this.generate(
            level, random, pos, rootReplacer, trunkReplacer, foliageReplacer, config
        );
        if (!generate || trunkPos.isEmpty() && foliagePos.isEmpty())
            return false;
        if (!config.field_21290.isEmpty()) {
            ArrayList<class_2338> rootPositions = Lists.newArrayList(rootPos);
            ArrayList<class_2338> trunkPositions = Lists.newArrayList(trunkPos);
            ArrayList<class_2338> foliagePositions = Lists.newArrayList(foliagePos);
            trunkPositions.sort(Comparator.comparingInt(class_2382::method_10264));
            foliagePositions.sort(Comparator.comparingInt(class_2382::method_10264));
            rootPositions.sort(Comparator.comparingInt(class_2382::method_10264));
            config.field_21290.forEach(
                treeDecorator
                -> treeDecorator.method_23469(
                    level, decoratorReplacer, random, trunkPositions, foliagePositions
                )
            );
        }

        return class_3341
            .method_35411(Iterables.concat(trunkPos, foliagePos, decoratorPos))
            .map(box -> {
                class_251 shape = RootedTreeFeature.placeLogsAndLeaves(
                    level, box, trunkPos, decoratorPos
                );
                class_3499.method_20532(
                    level, class_2248.field_31036, shape, box.method_35415(), box.method_35416(), box.method_35417()
                );
                return true;
            })
            .orElse(false);
    }

    private static class_251 placeLogsAndLeaves(
        class_1936 level,
        class_3341 box,
        Set<class_2338> trunkPositions,
        Set<class_2338> decoratorPositions
    ) {
        ArrayList<Set<class_2338>> positions = Lists.newArrayList();
        class_244 shape = new class_244(
            box.method_35414(), box.method_14660(), box.method_14663()
        );

        for (int tries = 0; tries < 6; ++tries)
            positions.add(Sets.newHashSet());

        class_2338.class_2339 mutable = new class_2338.class_2339();
        for (class_2338 pos : Lists.newArrayList(decoratorPositions))
            if (box.method_14662(pos))
                shape.method_1049(
                    pos.method_10263() - box.method_35415(),
                    pos.method_10264() - box.method_35416(),
                    pos.method_10260() - box.method_35417()
                );

        for (class_2338 pos : Lists.newArrayList(trunkPositions)) {
            if (box.method_14662(pos))
                shape.method_1049(
                    pos.method_10263() - box.method_35415(),
                    pos.method_10264() - box.method_35416(),
                    pos.method_10260() - box.method_35417()
                );

            for (class_2350 direction : class_2350.values()) {
                class_2680 blockState;
                mutable.method_25505(pos, direction);
                if (trunkPositions.contains(mutable)
                    || !(blockState = level.method_8320(mutable))
                            .method_28498(class_2741.field_12541))
                    continue;
                positions.get(0).add(mutable.method_10062());
                TreeFeatureAccessor.setBlockKnownShape(
                    level, mutable, blockState.method_11657(class_2741.field_12541, 1)
                );
                if (!box.method_14662(mutable))
                    continue;
                shape.method_1049(
                    mutable.method_10263() - box.method_35415(),
                    mutable.method_10264() - box.method_35416(),
                    mutable.method_10260() - box.method_35417()
                );
            }
        }

        for (int tries = 1; tries < 6; ++tries) {
            Set<class_2338> trunkPos = positions.get(tries - 1);
            Set<class_2338> foliagePos = positions.get(tries);
            for (class_2338 pos : trunkPos) {
                if (box.method_14662(pos))
                    shape.method_1049(
                        pos.method_10263() - box.method_35415(),
                        pos.method_10264() - box.method_35416(),
                        pos.method_10260() - box.method_35417()
                    );

                for (class_2350 direction : class_2350.values()) {
                    mutable.method_25505(pos, direction);
                    class_2680 state = level.method_8320(mutable);
                    if (trunkPos.contains(mutable) || foliagePos.contains(mutable)
                        || !state.method_28498(class_2741.field_12541)
                        || (state.method_11654(class_2741.field_12541)) <= tries + 1)
                        continue;
                    class_2680 foliage
                        = state.method_11657(class_2741.field_12541, tries + 1);
                    TreeFeatureAccessor.setBlockKnownShape(level, mutable, foliage);
                    if (box.method_14662(mutable))
                        (shape).method_1049(
                            mutable.method_10263() - box.method_35415(),
                            mutable.method_10264() - box.method_35416(),
                            mutable.method_10260() - box.method_35417()
                        );
                    foliagePos.add(mutable.method_10062());
                }
            }
        }
        return shape;
    }
}